/* $Id: Board.java,v 1.26 2007/11/28 17:49:42 ekstrand Exp $ */

package csci1902.hw6.game;

import java.util.*;
import java.lang.ref.*;


/**
 * Class representing a game board.
 *
 * Game boards are immutable.  This is necessary for them to be map keys
 * (which is necessary for the memoization at the B level).  To make a play on
 * a board, create a new board based on the first board with a new move.
 * 
 * The {@link #hashCode} method is implemented meaningfully, so game boards
 * are valid keys in HashMap's.
 * 
 * @author Michael Ekstrand <ekstrand@cs.umn.edu>
 */
public class Board {
    /** The board's configuration */
    private final Config config;

    /** The internal representation of the board */
    private long board = 0L;

    /** A cache of available moves */
    private SoftReference<Map<CellPosition,Collection<CellPosition>>> availableMoves = null;
    
    /**
     * Create a 4x8 board for 4-in-a-row.
     */
    public Board() {
        this(4, 8, 4);
    }
    
    /**
     * Create a new blank board
     * @param rows The board's row count
     * @param columns The board's column count
     * @param goal The board's goal (target number in a row)
     */
    public Board(int rows, int columns, int goal) {
        config = new Config(rows, columns, goal);
    }
    
    /**
     * Create a new board by adding one play to a base board.
     * @param base The base board
     * @param player The player adding the play
     * @param move The play to add.
     */
    public Board(Board base, PlayerID player, CellPosition move) {
        config = base.config;
        board = base.board;
        if (getCellValue(move) != null)
            throw new BadMoveException("Invalid move");
        setCell(player, move);
    }

    /** Query the number of rows in the board.
     * @return the number of rows in this board
     */
    public int getRows() {
        return config.rowCount;
    }

    /** Query the number of columns in the board.
     * @return the number of columns in this board
     */
    public int getColumns() {
        return config.columnCount;
    }
    
    /**
     * Queries the goal (number in a row) for this game.
     * @return The goal for this board.
     */
    public int getGoal() {
        return config.goal;
    }
    
    /**
     * Obtains the list of win positions (sets of cells which comprise a win when
     * all are occupied by the same player).  The returned list must not be
     * modified.
     * @return The list of win position sequences.
     */
    public List<CellPosition[]> getWinPositions() {
        return Collections.unmodifiableList(config.winPositions);
    }
    
    /**
     * Obtain a CellPosition object representing a specified row and column.
     * @param row The cell's row.
     * @param column The cell's column.
     * @return A CellPosition representing this cell.
     */
    public CellPosition makeCellPosition(int row, int column) {
        return config.makeCellPosition(row, column);
    }
    
    /**
     * Retrieve the value of the cell at a certain position.
     * @param cell The position to query
     * @return The player whose token is at that cell (or null, if the cell is
     * empty).
     */
    public PlayerID getCellValue(CellPosition cell) {
        long sval = getShiftValue(cell);
        long mask = 3L << sval;
        long bmasked = board & mask;        
        long code = bmasked >>> sval;
        assert code <= 3 && code >= 0;
        return numToId(code); 
    }
    
    /**
     * Retrieve the cell value at a particular row and column.
     * @param row The row to query.
     * @param column The column to query.
     * @return The cell value at the specified location ({@code null} if no play
     * has been made at that position).
     */
    public PlayerID getCellValue(int row, int column) {
        return getCellValue(makeCellPosition(row, column));
    }
    
    /**
     * Query the winner of the board.
     * @return The winner, or {@code null} if there is no winner.
     */
    public PlayerID getWinner() {
        PlayerID winner = null;
        
        for (CellPosition[] pos: config.winPositions) {
            for (int i = 0; i < pos.length; i++) {
                CellPosition p = pos[i];
                PlayerID cell = getCellValue(p);
                if (i == 0) {
                    winner = cell;
                    if (winner == null)
                        break;
                } else if (!winner.equals(cell)) {
                    winner = null;
                    break;
                }
            }
            if (winner != null)
                break;
        }
        
        return winner;
    }
    
    /**
     * Query whether the board is a draw (no more moves left).
     * @return {@code true} of the game is a draw.
     */
    public boolean isDraw() {
        if (getWinner() != null)
            return false;
        
        for (int r = 0; r < config.rowCount; r++) {
            for (int c = 0; c < config.columnCount; c++) {
                // TODO: improve efficiency
                if (getCellValue(r,c) == null)
                    return false;
            }
        }
        
        return true;
    }
    
    /**
     * Query whether the board represents a finished game.
     * @return {@code true} if the game is over - there is either a winner or a
     * draw.
     */
    public boolean isFinished() {
        return getWinner() != null || isDraw();
    }
    
    /**
     * Set the value of a particular player's cell.
     * @param player The player to play at the position.
     * @param position The cell position to play.
     */
    private void setCell(PlayerID player, CellPosition position) {
        int sval = getShiftValue(position);
        long cval = idToNum(player);
        board &= ~(3L << sval);
        board |= cval << sval;
    }
    
    /**
     * Convert a cell position to a shift value.
     * @see #getShiftValue(int, int)
     * @param position The position to convert.
     * @return The position's shift value.
     */
    private int getShiftValue(CellPosition position) {
        return getShiftValue(position.row, position.column);
    }
    
    /**
     * Convert a cell position to a shift value.
     * @param row The row to convert.
     * @param column The column to convert.
     * @return The shift value required to get to the cell's position in the
     * board data.
     */
    private int getShiftValue(int row, int column) {
        return column * 8 + row * 2;
    }
    
    /**
     * Convert a player encoding (number) to a player ID.
     * @param num The number to convert.
     * @return The ID of that player, or null if it represents an empty square.
     */
    private PlayerID numToId(long num) {
        switch ((int) num) {
        case 0:
            return null;
        case 1:
            return PlayerID.X;
        case 2:
            return PlayerID.O;
        default:
            throw new RuntimeException("Invalid player code");
        }
    }
    /**
     * Convert a {@link PlayerID} to a number for the game board.
     * @param player The player ID to convert.
     * @return The encoding of that player for the game board.
     */
    private int idToNum(PlayerID player) {
        if (player == null) {
            return 0;
        } else {
            switch (player) {
            case X: return 1;
            case O: return 2;
            }
        }
        throw new RuntimeException("Invalid player");
    }
    
    /**
     * Retrieve a list/collection of all symmetrically unique available moves
     * (open spaces) in the game board.
     * @see #symmetricallyEqual(long, long)
     * @return A collection of all available moves.
     */
    public Collection<CellPosition> getAvailableMoves() {
        return getAvailableMoveEquivs().keySet();
    }
    
    /**
     * Get the equivalent moves to a particular move.
     * @param move The move whose equivalence class should be returned.
     * @return The collection of all moves which are symmetrically equal to
     * {@code move}, including {@code move} itself.
     */
    public Collection<CellPosition> getEquivalentMoves(CellPosition move) {
        Map<CellPosition,Collection<CellPosition>> mapping;
        mapping = getAvailableMoveEquivs();
        
        Collection<CellPosition> equivs = mapping.get(move);
        if (equivs == null) {
            equivs = new ArrayList<CellPosition>(1);
            equivs.add(move);
        }
        return equivs;
    }
    
    /**
     * Procure (from cache, if possible) the list of available moves with
     * equivalence classes.
     * @return The map of available moves to equivalence classes.
     */
    private Map<CellPosition,Collection<CellPosition>> getAvailableMoveEquivs() {
        Map<CellPosition,Collection<CellPosition>> mlist;
        if (availableMoves == null || availableMoves.get() == null) {
            mlist = getAvailableMoves(true);
            availableMoves =
                new SoftReference<
                Map<CellPosition,Collection<CellPosition>>>(mlist);
        } else {
            mlist = availableMoves.get();
        }
        return mlist;
    }
    
    /**
     * Return the list of avaialble moves in the board, optionally removing
     * symmetries.
     * @param removeSymmetries If true, restrict result to symmetrically unique
     * moves.
     * @return The available moves (open spaces) in the board.
     */
    private Map<CellPosition,Collection<CellPosition>> getAvailableMoves(boolean removeSymmetries) {
        if (removeSymmetries) {
            if (board == 0) {
                int ncells = (config.rowCount + 1) / 2;
                Map<CellPosition,Collection<CellPosition>> cells =
                    new TreeMap<CellPosition,Collection<CellPosition>>();
                for (int i = 0; i < ncells; i++) {
                    CellPosition p = makeCellPosition(i, 0);
                    List<CellPosition> equivs = new ArrayList<CellPosition>();
                    cells.put(p, equivs);
                    // add rotations
                    for (int j = 0; j < config.columnCount; j++)
                        equivs.add(makeCellPosition(i, j));
                    // vertically flip, add rotations
                    int ni = config.rowCount - i;
                    if (ni != i) {
                        for (int j = 0; j < config.columnCount; j++)
                            equivs.add(makeCellPosition(ni, j));
                    }
                }
                return cells;
            } else {
                int nbits = 0;
                for (int i = 0; i < 64; i++) {
                    if ((board & (1L << i)) != 0)
                        nbits++;
                }
                if (nbits > 8)
                    // don't scan for symmetry on large boards
                    return getAvailableMoves(false);
                else
                    return filterSymmetries(getAvailableMoves(false).keySet());
            }
        } else {
            Map<CellPosition,Collection<CellPosition>> moves =
                new TreeMap<CellPosition,Collection<CellPosition>>();
            for (int r = 0; r < config.rowCount; r++) {
                for (int c = 0; c < config.columnCount; c++) {
                    CellPosition pos = makeCellPosition(r, c);
                    if (getCellValue(pos) == null)
                        moves.put(pos, null);
                }
            }
            return moves;
        }
    }
    
    /**
     * Filter out the symmetrical moves from a list of cell positions.  Two
     * moves are symmetrical if they result in symmetrical boards.
     * @see #symmetricallyEqual(long, long)
     * @param cells The list of cells to examine.
     * @return The list of unique moves.
     */
    private Map<CellPosition,Collection<CellPosition>>
    filterSymmetries(Collection<CellPosition> cells) {
        Map<CellPosition,Collection<CellPosition>> equivs =
            new TreeMap<CellPosition,Collection<CellPosition>>();
        Map<CellPosition,Long> filtered = new TreeMap<CellPosition,Long>();
        
        for (CellPosition cell: cells) {
            int sval = getShiftValue(cell);
            long grid = board | (3L << sval);
            boolean found = false;
            CellPosition equiv = null;
            for (Map.Entry<CellPosition, Long> entry: filtered.entrySet()) {
                if (symmetricallyEqual(grid, entry.getValue())) {
                    equiv = entry.getKey();
                    break;
                }
            }
            if (equiv != null) {
                equivs.get(equiv).add(cell);
            } else {
                filtered.put(cell, grid);
                List<CellPosition> lst = new ArrayList<CellPosition>();
                lst.add(cell);
                equivs.put(cell, lst);
            }
        }
        
        return equivs;
    }
    
    /**
     * Check two boards for equality.
     */
    @Override
    public boolean equals(Object other) {
        if (other instanceof Board) {
            Board oboard = (Board) other;
            
            return board == oboard.board;
        } else {
            return false;
        }
    }
    
    /**
     * Check two grids for symmetric equality.  Two boards are symmetrically
     * equal if one can be obtained by rotating and/or flipping the other.  This
     * method does not consider vertical flips; they only matter with an empty
     * board, and are handled as a special case by
     * {@link #filterSymmetries(Collection)}(filterSymmetries).
     * @param b1 One board grid to compare
     * @param b2 The other board grid to compare
     * @return true if b1 and b2 are equal, considering symmetric boards to be
     * equal.
     */
    private boolean symmetricallyEqual(long b1, long b2) {
        if (b1 == b2)
            return true;

        long adjusted = b1;

        // scan rotations
        for (int i = 0; i < 7; i++) {
            adjusted = rotate(adjusted);
            if (b2 == adjusted)
                return true;
        }

        // flip, scan again
        adjusted = flipHorizontally(b1);
        if (b2 == adjusted)
            return true;

        for (int i = 0; i < 7; i++) {
            adjusted = rotate(adjusted);
            if (b2 == adjusted)
                return true;
        }

        // we got nothing
        return false;
    }
    
    /**
     * Rotate a grid by columns.
     * @param grid The grid to rotate.
     * @return {@code grid}, rotated by one column.
     */
    private long rotate(long grid) {
        long rboard = 0L;
        for (int i = 0; i < config.columnCount; i++) {
            int ri = (i + 1) % config.columnCount;
            rboard |= ((grid >>> (i * 8)) & 0xFFL) << (ri * 8);
        }
        return rboard;
    }
    
    /**
     * Flip a grid horizontally.
     * @param grid The grid to flip.
     * @return The flipped grid.
     */
    private long flipHorizontally(long grid) {
        long nboard = 0L;
        for (int c = 0; c < config.columnCount; c++) {
            int cinv = config.columnCount - c - 1;
            long data = grid >>> (c * 8);
            data &= 0xFFL;
            nboard |= data << (cinv * 8);
        }
        return nboard;
    }
    
    /**
     * Compute the hash code of a board for hash table/HashMap purposes.  The
     * hash is the same as is used by {@link java.lang.Long#hashCode()}.
     */
    @Override
    public int hashCode() {
        return (int) (board ^ (board >>> 32));
    }
    
    /**
     * Internal class for representing and sharing board configurations.
     * @author Michael Ekstrand <ekstrand@cs.umn.edu>
     *
     */
    static class Config {
        /** The number of rows in a board. */
        public final int rowCount;
        /** The number of columns in a board. */
        public final int columnCount;
        /** The goal of a board &emdash; the number of cells to get in a row */
        public final int goal;
        /** The list of winning position sequences in the board. */
        public final List<CellPosition[]> winPositions;
        
        /**
         * Initialize a new board configuration.
         * @param rows The rows in the configuration.
         * @param columns The columns in the configuration.
         * @param goal The goal of the configuration.
         */
        public Config(int rows, int columns, int goal) {
            rowCount = rows;
            if (rowCount > 4)
                throw new IndexOutOfBoundsException("Invalid board height: " + rows);
            columnCount = columns;
            if (rowCount > 8)
                throw new IndexOutOfBoundsException("Invalid board width: " + columns);
            this.goal = goal;
            winPositions = computeWinPositions();
        }

        /**
         * Create a new {@link CellPosition} with this configuration.
         */
        public CellPosition makeCellPosition(int row, int column) {
            return new CellPosition(this, row, column);
        }
        
        /**
         * Compute the list of all winning position sequences in the board.
         * @return The list of all cell position sequences with which the board
         * can be won.
         */
        private List<CellPosition[]> computeWinPositions() {
            ArrayList<CellPosition[]> posns = new ArrayList<CellPosition[]>(rowCount * columnCount * 4);
            for (int r = 0; r < rowCount; r ++) {
                for (int c = 0; c < columnCount; c++) {
                    CellPosition[] pos = new CellPosition[goal];
                    for (int i = 0; i < goal; i++)
                        pos[i] = makeCellPosition(r, c+i);
                    posns.add(pos);
                    
                    if (r+goal <= rowCount) {
                        pos = new CellPosition[goal];
                        for (int i = 0; i < goal; i++)
                            pos[i] = makeCellPosition(r+i, c);
                        posns.add(pos);
                        pos = new CellPosition[goal];
                        for (int i = 0; i < goal; i++)
                            pos[i] = makeCellPosition(r+i, c+i);
                        posns.add(pos);
                        pos = new CellPosition[goal];
                        for (int i = 0; i < goal; i++)
                            pos[i] = makeCellPosition(r+i,c-i);
                        posns.add(pos);
                    }
                }
            }
            posns.trimToSize();
            return Collections.unmodifiableList(posns);
        }
    }  
}
