/* $Id: Player.java,v 1.4 2007/11/28 17:49:42 ekstrand Exp $ */

package csci1902.hw6.game;

import java.util.*;
import java.util.logging.Logger;
import javax.swing.event.ChangeListener;
import javax.swing.event.ChangeEvent;

/**
 * Abstract class for implementing players (entities which can play the game
 * of tic-tac-toe).
 *
 * <h2>Asynchronous Interface</h2>
 * The interface between the game and players is asynchronous.  The game
 * requests a move from a player using the {@link #requestMove(MoveListener)}
 * method.  The player then responds with the move by invoking the {@link
 * MoveListener#move(CellPosition)} method on the listener with the position
 * moved to.  This response can happen after requestMove terminates.
 *
 * @author Michael Ekstrand <ekstrand@cs.umn.edu>
 */
public abstract class Player {
    private static Logger logger = Logger.getLogger("csci1902.hw6.Game");

    /** The ID of the player. */
    protected final PlayerID id;
    /** The game being played. */
    protected Game game;
    /** The current move listener. */
    private MoveListener listener = null;
    /** The list of listeners interested in player changes. */
    private List<ChangeListener> changeListeners;

    /**
     * Construct a new Player (overridden by derived classes).
     */
    protected Player(PlayerID player) {
        changeListeners = new LinkedList<ChangeListener>();
        this.id = player;
        this.game = null;
    }

    /**
     * Query the ID of the player.
     * @return The player's ID
     */
    public PlayerID getId() {
        return id;
    }

    /**
     * Set the game to be played.
     * @note This should only called once.
     * @param game The game to play.
     */
    public void setGame(Game game) {
        if (this.game != null)
            throw new RuntimeException("Attempted to re-use player");
        this.game = game;
    }

    /**
     * Inform this player of an opponent's move.  The default implementation
     * does nothing; specific player subclasses should override this if they
     * want to do something with opponent moves.
     * @param opponent The player who moved.
     * @param pos The position the player moved to.
     */
    public void postOpponentMove(PlayerID opponent, CellPosition pos) {
        
    }

    /**
     * Query whether this player is active (it's this player's turn).
     */
    public boolean isActive() {
        return listener != null;
    }

    /**
     * Request a move from the player.  Subclasses should override this, chain
     * up to the base class implementation, and do something useful.
     * @param listener The listener which wants to be notified of this move.
     */
    public void requestMove(MoveListener listener) {
        logger.fine("Move requested from player " + id);
        this.listener = listener;
        notifyChange();
    }

    /**
     * Notify the listener of a move.  This class informs the current listener
     * ({@link #listener}) of a move, and notifies change listeners that the
     * player has changed.  It also clears the listener.
     * @param move The move being made.
     */
    protected void notifyMove(CellPosition move) {
        if (listener != null) {
            listener.move(move);
            listener = null;
            notifyChange();
        } else {
            throw new RuntimeException("Cannot post move without listener");
        }
    }

    /**
     * Register a listener to be notified when the player changes state.
     * @param l The listener to add.
     */
    public void addChangeListener(ChangeListener l) {
        changeListeners.add(l);
    }
    /**
     * Unregister a listener to be notified when the player changes state.
     * @param l The listener to remove.
     */
    public void removeChangeListener(ChangeListener l) {
        changeListeners.remove(l);
    }
    /**
     * Notify all interested listeners that the player has changed state.
     */
    protected void notifyChange() {
        ChangeEvent evt = new ChangeEvent(this);
        for (ChangeListener l: changeListeners) {
            l.stateChanged(evt);
        }
    }

}
