/* $Id: GamePanel.java,v 1.12 2007/11/28 17:49:43 ekstrand Exp $ */

package csci1902.hw6.ui;

import java.awt.*;
import java.awt.event.*;
import java.util.List;
import java.util.LinkedList;
import java.util.Map;
import java.util.TreeMap;
import java.util.logging.Logger;
import javax.swing.*;
import javax.swing.event.*;

import static java.lang.Math.round;

import csci1902.hw6.game.Board;
import csci1902.hw6.game.CellPosition;
import csci1902.hw6.game.Game;
import csci1902.hw6.game.PlayerID;

@SuppressWarnings("serial")
class GamePanel extends JPanel implements ChangeListener {
    private static int BUFFER = 10;
    
    private static Color X_COLOR = new Color(0.0f, 0.0f, 0.8f);
    private static Color O_COLOR = new Color(0.8f, 0.0f, 0.0f);
    
    private static Logger logger = Logger.getLogger("csci1902.hw6.ui.GamePanel");
    
    private Game game;
    private Dimension size;
    private int dim, space;
    private double cellAngle;
    private int numCols, numRows;
    private List<CellClickListener> cellClickListeners;
    private Map<CellPosition,String> cellReports;
    
    public GamePanel() {
        cellClickListeners = new LinkedList<CellClickListener>();
        
        Dimension msize = new Dimension(400, 400);
        setMinimumSize(msize);
        setPreferredSize(msize);
        
        addMouseListener(new MouseHandler());
        
        cellReports = new TreeMap<CellPosition,String>();
    }
    
    public void setGame(Game game) {
        if (this.game != null)
            this.game.removeChangeListener(this);
        this.game = game;
        game.addChangeListener(this);
        cellReports.clear();
        repaint();
    }
    
    public void paintComponent(Graphics g) {
        if (game == null)
            return;
        Graphics2D g2 = (Graphics2D) g;
        updateSizes();
        
        g2.clearRect(0, 0, size.width, size.height);
        
        paintGrid(g2);
        paintBoard(g2);
    }
    
    private void updateSizes() {
        if (game != null) {
            numCols = game.getBoard().getColumns();
            numRows = game.getBoard().getRows();
        } else {
            numCols = 8;
            numRows = 4;
        }
        
        size = getSize();
        logger.fine(String.format("Board has width=%d, height=%d",
                size.width, size.height));
        dim = (size.width < size.height ? size.width : size.height) - BUFFER * 2;
        space = dim / ((numRows + 1) * 2);
        logger.fine(String.format("dim=%d, space=%d", dim, space));
        
        cellAngle = Math.PI * 2 / numCols;
    }
    
    private void paintGrid(Graphics2D g) {
        int center_x = size.width / 2;
        int center_y = size.height / 2;
        
        int off_x = (size.width - dim) / 2;
        int off_y = (size.height - dim) / 2;
        
        for (int i = 0; i < numRows + 1; i++) {
            int x = off_x + space * i;
            int y = off_y + space * i;
            int d = dim - space * i * 2;
            if (i < numRows)
                g.drawOval(x, y, d, d);
            else
                g.fillOval(x, y, d, d);
        }
        
        for (int i = 0; i < numCols; i++) {
            double rad = cellAngle * i;
            g.drawLine(center_x, center_y,
                    center_x + (int) round(dim / 2 * Math.cos(rad)),
                    center_y + (int) round(dim / 2 * Math.sin(rad)));
        }
    }
    
    private void paintBoard(Graphics2D g2) {
        Board board = game.getBoard();
        for (int r = 0; r < board.getRows(); r++) {
            for (int c = 0; c < board.getColumns(); c++) {
                PlayerID player = board.getCellValue(r,c);
                if (player != null) {
                    paintCell(g2, r, c, player);
                }
            }
        }
    }
    
    private void paintCell(Graphics2D g, int r, int c, PlayerID player) {
        double th = cellAngle * (c + 0.5);
        double rcent = space * 3 / 2.0 + r * space;
        
        double x = rcent * Math.cos(th);
        double y = rcent * Math.sin(th);
        
        logger.finer(String.format("Painting %s at %d,%d (x=%f,y=%f) (r=%f,th=%f)",
                player, r, c, x, y, rcent, th));
        
        x += size.width / 2;
        y += size.height / 2;
        
        switch (player) {
        case X:
            paintX(g, (int) x, (int) y);
            break;
        case O:
            paintO(g, (int) x, (int) y);
            break;
        }
    }
    
    private void paintX(Graphics2D g, int x, int y) {
        Graphics2D g2 = (Graphics2D) g.create();
        g2.setPaint(X_COLOR);
        double rsize = space / 3.0;
        g2.drawLine((int) round(x-(rsize/2)), (int) round(y-(rsize/2)),
                (int) round(x+(rsize/2)), (int) round(y+(rsize/2)));
        g2.drawLine((int) round(x-(rsize/2)), (int) round(y+(rsize/2)),
                (int) round(x+(rsize/2)), (int) round(y-(rsize/2)));
    }
    
    private void paintO(Graphics2D g, int x, int y) {
        Graphics2D g2 = (Graphics2D) g.create();
        g2.setPaint(O_COLOR);
        
        double rsize = space / 3.0;
        
        g2.drawOval((int) (x-(rsize/2)), (int) (y-(rsize/2)), (int) rsize, (int) rsize);
    }

    public void stateChanged(ChangeEvent arg0) {
        cellReports.clear();
        repaint();
    }
    
    private void onMouseClicked(int x, int y) {
        logger.fine(String.format("Click received: x=%d, y=%d", x, y));
        CellPosition cell = posToCell(x, y);
        if (cell != null) {
            logger.fine("Dispatching as cell " + cell);
            notifyCellClicked(cell);
        }
    }
    
    private CellPosition posToCell(int x, int y) {
        if (game == null)
            return null;
        
        updateSizes();
        x -= size.width / 2;
        y -= size.height / 2;
        double th = Math.atan2(y, x);   // this should be x,y; not sure why this works
        double r = Math.hypot(x, y);
        
        if (r < dim && r > space) {
            int row = (int) (r / space) - 1;
            if (row >= game.getBoard().getRows())
                return null;
            if (th < 0)
                th += 2 * Math.PI;
            int col = (int) (th / (2 * Math.PI) * numCols) % numCols;            
            return game.getBoard().makeCellPosition(row, col);
        } else {
            return null;
        }
    }
    
    public void addCellClickListener(CellClickListener l) {
        cellClickListeners.add(l);
    }
    public void removeCellClickListener(CellClickListener l) {
        cellClickListeners.remove(l);
    }
    private void notifyCellClicked(CellPosition pos) {
        CellClickEvent evt = new CellClickEvent(this, pos);
        for (CellClickListener l: cellClickListeners) {
            l.cellClicked(evt);
        }
    }
    
    @SuppressWarnings("unused")
    private class MouseHandler extends MouseAdapter {
        @Override
        public void mouseClicked(MouseEvent event) {
            if (event.getButton() == 1)
                onMouseClicked(event.getX(), event.getY());
        }
    }
}
