/* $Id: MainWindow.java,v 1.24 2007/11/28 19:39:28 ekstrand Exp $ */

/**
 * The user interface for the tic-tac-toe game system.
 */
package csci1902.hw6.ui;

import java.awt.*;
import java.awt.event.*;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.util.logging.Logger;
import javax.swing.*;
import javax.swing.event.*;

import csci1902.hw6.ai.ComputerPlayer;
import csci1902.hw6.ai.BoardHeuristic;
import csci1902.hw6.ai.SimpleHeuristic;

import csci1902.hw6.game.Board;
import csci1902.hw6.game.Game;
import csci1902.hw6.game.Player;
import csci1902.hw6.game.PlayerID;

/**
 * Main window of the tic-tac-toe game.
 * @author Michael Ekstrand <ekstrand@cs.umn.edu>
 *
 */
@SuppressWarnings("serial")
public class MainWindow
    extends JFrame
    implements ActionListener, ChangeListener {
    
    private static Logger logger = Logger.getLogger("csci1902.hw6.ui");

    /**
     * The main entry point to run tic-tac-toe.
     * @param args The command-line arguments.
     */
    public static void main(String[] args) {
        MainWindow window = new MainWindow();
        window.setVisible(true);
        logger.info("GUI launched");
    }

    private Player player1, player2;
    private Game game;
    private BoardHeuristic simpleHeuristic = null;
    private BoardHeuristic improvedHeuristic = null;
    
    private Timer updateTimer;
    private JLabel messageLabel, memStatus;
    private long lastStartTime = -1;
    private GamePanel gamePanel;
    
    private JDialog newGameDialog;
    private ButtonGroup player1Buttons, player2Buttons, gameTypeButtons;

    /**
     * Construct a new main window.
     */
    @SuppressWarnings("unchecked")
    public MainWindow() {
        super();

        simpleHeuristic = new SimpleHeuristic();
        
        try {
            Class cls = Class.forName("csci1902.hw6.ai.ImprovedHeuristic");
            Constructor ctor = cls.getConstructor();
            improvedHeuristic = (BoardHeuristic) ctor.newInstance();
        } catch (ClassNotFoundException e) {
            logger.warning("Improved AI heuristic not available");
        } catch (ClassCastException e) {
            logger.warning("Improved heuristic invalid (not BoardHeuristic)");
        } catch (NoSuchMethodException e) {
            logger.warning("Improved heuristic does not provide default constructor");
        } catch (InstantiationException e) {
            logger.warning("Cannot instantiate improved heuristic");
        } catch (IllegalAccessException e) {
            logger.warning("Cannot instantiate improved heuristic (access violation)");
        } catch (InvocationTargetException e) {
            logger.warning("Error instantiating improved heuristic");
        }
        
        setDefaultCloseOperation(EXIT_ON_CLOSE);
        setLocationRelativeTo(null);
        setLocationByPlatform(true);
        setTitle("Tic-Tac-Toe");
        setLayout(new BorderLayout());
        
        JToolBar toolBar = new JToolBar();
        add(toolBar, BorderLayout.NORTH);
        
        JButton newGame = new JButton("New...");
        newGame.setActionCommand("newGame");
        newGame.addActionListener(this);
        toolBar.add(newGame);
        
        toolBar.add(new JSeparator());
        
        JButton quit = new JButton("Quit");
        toolBar.add(quit);
        quit.setActionCommand("quit");
        quit.addActionListener(this);
        
        messageLabel = new JLabel();
        memStatus = new JLabel();
        memStatus.setHorizontalAlignment(SwingConstants.RIGHT);
        Container sbox = new Box(BoxLayout.X_AXIS);
        sbox.add(messageLabel);
        sbox.add(Box.createGlue());
        sbox.add(memStatus);
        add(sbox, BorderLayout.SOUTH);

        doPeriodicUpdates();
        
        updateTimer = new Timer(500, this);
        updateTimer.start();
        
        gamePanel = new GamePanel();
        add(gamePanel, BorderLayout.CENTER);
        
        pack();
    }

    /**
     * Respond to an action (button click, etc.).
     */
    public void actionPerformed(ActionEvent evt) {
        String cmd = evt.getActionCommand();
        
        if (evt.getSource() == updateTimer) {
            doPeriodicUpdates();
        } else if (cmd.equals("newGame")) {
            if (newGameDialog == null)
                buildNewGameDialog();

            newGameDialog.setVisible(true);
        } else if (cmd.equals("quit")) {
            System.exit(0);
        } else if (cmd.equals("newGameCancel")) {
            newGameDialog.setVisible(false);
        } else if (cmd.equals("newGameAccept")) {
            newGameDialog.setVisible(false);
            startNewGame();
        } else {
            logger.warning("Invalid command: " + cmd);
        }
    }

    /**
     * Respond to a state change, either by the game or a player.
     */
    public void stateChanged(ChangeEvent evt) {
        Object src = evt.getSource();
        if (src == player1) {
            if (player1.isActive()) {
                if (player1 instanceof GuiPlayer)
                    messageLabel.setText("Player 1's turn");
                else
                    messageLabel.setText("Computer player 1 moving...");
                lastStartTime = System.currentTimeMillis();
            }
        } else if (src == player2) {
            if (player2.isActive()) {
                if (player2 instanceof GuiPlayer)
                    messageLabel.setText("Player 2's turn");
                else
                    messageLabel.setText("Computer player 2 moving...");
                lastStartTime = System.currentTimeMillis();
            }
        } else if (src == game) {
            Board b = game.getBoard();
            if (b.isFinished()) {
                String msg;
                if (b.isDraw()) {
                    msg = "Draw!";
                } else if (b.getWinner().equals(player1.getId())) {
                    msg = "Player 1 won!";
                } else {
                    msg = "Player 2 won!";
                }
                messageLabel.setText(msg);
                lastStartTime = -1;
            }
        }
    }

    /**
     * Construct a dialog for new game parameters.
     */
    private void buildNewGameDialog() {
        newGameDialog = new JDialog(this, "New Game", true);
        newGameDialog.setLocationRelativeTo(this);
        newGameDialog.setLocationByPlatform(true);
        LayoutManager layout =
            new BoxLayout(newGameDialog.getContentPane(),
                          BoxLayout.Y_AXIS);
        newGameDialog.setLayout(layout);
        
        gameTypeButtons = new ButtonGroup();
        JPanel gtpanel = new JPanel();
        gtpanel.setLayout(new BoxLayout(gtpanel, BoxLayout.Y_AXIS));
        gtpanel.setBorder(BorderFactory.createTitledBorder("Game type"));
        JRadioButton rb = new JRadioButton("3x6 (3 in a row)");
        rb.setActionCommand("36");
        gameTypeButtons.add(rb);
        gtpanel.add(rb);
        rb = new JRadioButton("4x8 (4 in a row)");
        rb.setActionCommand("48");
        gameTypeButtons.add(rb);
        gtpanel.add(rb);
        rb.setSelected(true);
        newGameDialog.add(gtpanel);
        newGameDialog.add(Box.createVerticalStrut(5));
        
        player1Buttons = new ButtonGroup();
        newGameDialog.add(buildPlayerPanel("Player 1 (X)", player1Buttons));
        newGameDialog.add(Box.createVerticalStrut(5));
        player2Buttons = new ButtonGroup();
        newGameDialog.add(buildPlayerPanel("Player 2 (O)", player2Buttons));
        
        newGameDialog.add(Box.createVerticalStrut(5));
        
        Box buttons = new Box(BoxLayout.X_AXIS);
        buttons.add(Box.createGlue());
        newGameDialog.add(buttons);
        
        JButton cancel = new JButton("Cancel");
        cancel.setActionCommand("newGameCancel");
        cancel.addActionListener(this);
        buttons.add(cancel);
        
        JButton ok = new JButton("OK");
        ok.setActionCommand("newGameAccept");
        ok.addActionListener(this);
        buttons.add(ok);
        
        newGameDialog.pack();
    }

    /**
     * Builds a new player panel for the new-game dialog.
     * @param title The panel's title.
     * @param buttons The button group into which the panel's radio buttons
     * should be inserted.
     * @return The panel containing the player options.
     */
    private JPanel buildPlayerPanel(String title, ButtonGroup buttons) {
        JPanel panel = new JPanel();
        panel.setBorder(BorderFactory.createTitledBorder(title));
        panel.setLayout(new BoxLayout(panel, BoxLayout.Y_AXIS));
        
        JRadioButton rb = new JRadioButton("Human");
        rb.setActionCommand("human");
        panel.add(rb);
        buttons.add(rb);
        rb.setSelected(true);
        
        rb = new JRadioButton("Computer (Simple)");
        rb.setActionCommand("simple");
        panel.add(rb);
        buttons.add(rb);
        rb.setEnabled(simpleHeuristic != null);
        
        rb = new JRadioButton("Computer (Improved)");
        rb.setActionCommand("improved");
        panel.add(rb);
        buttons.add(rb);
        rb.setEnabled(improvedHeuristic != null);
        
        return panel;
        
    }
    
    /**
     * Detach a player from its listeners.  It detaches this as a change
     * listener, and detaches the player from the game panel if appropriate.
     * @player The player to detach.
     */
    private void detachPlayer(Player player) {
        if (player != null) {
            player.removeChangeListener(this);
            if (player instanceof GuiPlayer)
                gamePanel.removeCellClickListener((GuiPlayer) player);
            player.removeChangeListener(this);
        }
    }

    /**
     * Do periodic updates of display things.
     */
    private void doPeriodicUpdates() {
        Runtime rt = Runtime.getRuntime();
        long tmem = rt.totalMemory() >> 20;
        long mmem = rt.maxMemory() >> 20;
        long fmem = rt.freeMemory() >> 20;
        long ctime = System.currentTimeMillis();
        String text =
            String.format("<html>Time: %d<br>Mem: %d/%d/%d</html>",
                          (lastStartTime >= 0)
                          ? ((ctime - lastStartTime) / 1000)
                          : 0,
                          tmem - fmem, tmem, mmem);
        memStatus.setText(text);
    }

    /**
     * Construct a new player.
     * @param id The player's ID.
     * @param spec The player's spec (either "human", "simple", or
     * "improved").
     */
    private Player makePlayer(PlayerID id, String spec) {
        Player player = null;
        if (spec.equals("human")) {
            GuiPlayer gp = new GuiPlayer(id);
            gamePanel.addCellClickListener(gp);
            player = gp;
        } else if (spec.equals("simple")) {
            player = new ComputerPlayer(simpleHeuristic, id, 4);
        } else if (spec.equals("improved")) {
            player = new ComputerPlayer(improvedHeuristic, id, 4);
        } else {
            logger.warning("Unknown player specification " + spec);
            throw new RuntimeException("Invalid player specification");
        }
        player.addChangeListener(this);
        return player;
    }

    /**
     * Start a new game by detaching current players and rebuilding
     * everything.
     */
    private void startNewGame() {
        detachPlayer(player1);
        player1 = null;
        detachPlayer(player2);
        player2 = null;
        
        player1 = makePlayer(PlayerID.X,
                player1Buttons.getSelection().getActionCommand());
        player2 = makePlayer(PlayerID.O,
                player2Buttons.getSelection().getActionCommand());
        
        String gameSpec = gameTypeButtons.getSelection().getActionCommand();
        if (gameSpec.equals("36"))
            game = new Game(3, 6, 3, player1, player2);
        else if (gameSpec.equals("48"))
            game = new Game(4, 8, 4, player1, player2);
        else
            throw new RuntimeException("Unknown game specification");
        
        gamePanel.setGame(game);
        game.addChangeListener(this);
        game.start();
    }

}
